#include "Global.H"
#include "GfxGlobal.H"
#include "Options.H"
#include "GfxModule.H"
#include "GameStateModule.H"
#include "ShutDown.H"
#include "SDLinit.H"
#include "Game.H"
#include "SquadCombatAI.H"
#include "ClientEventHandler.H"

using namespace std;

static char * VERSION = "BlackRose v1.0";

/* Tamaño de la ventana en modo gfx */
static const sint4 GFX_X = 900;  
static const sint4 GFX_Y = 740;

typedef SimpleTerrain::ST_ForceField Terrain;

/* Funcion para añadir las opciones del cliente */
void add_options(){
	Options opt("BlackRose");
	opt.put("-usegfx", "display world in 3d");	
}



/* Clase Looper */
class Looper : public LoopFunctor{

public:

  /* Constructor */
  Looper(ClientAIState * state_) : state(state_) {}

  /* Metodo loop */
  void loop() const {

	/* Si modo grafico */
    if (state->gfxm) {

      /* Si salir activado */
	  if (state->gfxm->get_quit()){
		state->quit = true;
		exit(0);
	  }

      bool refresh;
      Options::get("-refresh", refresh);

	  /* Si opcion de refresco activada */
      if (refresh) {

		/* Ratio maximo de refresco gfx */
        if (!state->just_drew)
          state->gfxm->draw();
        else
          state->just_drew = false;
      }
    }

	/* Busca mensajes del servidor */
    /* Si llega al menos uno, se llaman a los handlers */
    /* Si no, se duerme un milisegundo */
    if (!state->gsm->recv_view())
		SDL_Delay(1);
  }

private:
  ClientAIState * state;
};


#undef main
int main(int argc, char * * argv){

  /* Señales shut down activas */
  signals_shut_down(true);

  /* Añadimos todas las opciones, de los diferentes modulos */
  add_options();
  GameStateModule::Options::add();
  GfxModule::Options::add();
  Game::Options::add();
  Terrain::add_options(); // TODO: Fix this
  SquadCombatAI::add_options();
  
  ::Options::set("-nostartmotion", true);
  
  /* Procesar los argumentos de la linea de comandos */
  if (Options::process(argc, argv, cerr, VERSION))
    exit(20);

  /* Inicializamos SDL */
  SDLinit::video_init();
  SDLinit::network_init();

  /* Establecemos las opciones de los objetos con los valores de la linea de comandos y creamos modulos */
  GameStateModule::Options gsmo;
  GameStateModule gsm(gsmo);

  GfxModule::Options gfxmo;
  GfxModule * gfxm = 0;
  
  
  /* Cambiamos el valor de la semilla a time(0) si la opcion seed esta desactivada */
  sint4 seed;
  Options::get("-seed", seed);
  
  if (!seed)
	seed = time(0);
  
  Options::set("-seed", seed);  
  cout << "seed=" << seed << endl;
  
  SquadCombatAI scai(gsm);
  scai.enable_motion();
  scai.enable_attack();

  Watcher::WatcherModule wmGeneral(gsm);
  History::HistoryModule hmGeneral(gsm);
  TerrainAnalysis::Module tamGeneral(gsm, hmGeneral);
  Finance::Module fmGeneral(gsm, hmGeneral);
  
  Movement::Module::ptr mmGeneral(Movement::MakeModule(gsm, 110));
  mmGeneral->addPathfinder("Default", Movement::MakeSimpleTerrainPathfinder());
  mmGeneral->addPathExecutor("Default", Movement::MakeMultiFollowExecutor());	


  /* Mogo graficos 3d */
  bool use_gfx;
  Options::get("-usegfx", use_gfx);

  /* Si esta activado el modo grafico, se crea un modulo gfx */
  if (use_gfx)
	gfxm = new GfxModule;
	
 


  ModuleSet msGeneral;  
  msGeneral.set_movement(mmGeneral);  
  
  //msGeneral.set_watcher(&wmGeneral);
  //msGeneral.set_history(&hmGeneral);
  
  
  ModuleSet msGather;
  msGather.set_game_state(&gsm);
  msGather.set_combat(&scai);
  msGather.set_watcher(&wmGeneral);
  msGather.set_history(&hmGeneral);
  msGather.set_terrain_analysis(&tamGeneral);
  msGather.set_finance(&fmGeneral);
  msGather.set_movement(mmGeneral);
   
  /* Creamos una estructura ClienteAIState para el acceso al estado del juego por parte del handler */    
  //ClientAIState state(&gsm, gfxm, msGeneral, msGather);
  //ClientAIState state(&gsm, gfxm, msGeneral);
  ClientAIState state(&gsm, gfxm);
  
  /* Creamos un objeto ClienteEventHandler */
  ClientEventHandler sev(state, seed, msGeneral, msGather);

  /* Añadimos el handler al GameStateModule */
  gsm.add_handler(&sev); 
  
  gsm.add_handler(sev.simpleGather); 
  
  /* Conectamos el game state mudule con el servidor */
  if (!gsm.connect())
    ERR("connection problems");

  /* Creamos una instancia de la clase Looper */
  Looper l(&state);

  /* Si modo grafico 3d */
  if (use_gfx) {
 
	/* Inicializamos */   
    glutInit(&argc, argv);
    gfxm->init(gsm.get_game(), gsm.get_changes(), gsm.get_action_changes(), gfxmo);

    /* Llamamos al bucle del juego */
    gfxm->start_loop(&l);

  } else {

	/* Si modo no grafico llamamos al metodo loop hasta que tengamos que salir */
    while (!state.quit)
	  l.loop();
  }

  delete gfxm;
  
  return 0;
}