#ifndef ClientEventHandler_H
#define ClientEventHandler_H

#include "EventHandler.H"
#include "Random.H"
#include "Game.H"
#include "Movement.H"
#include "ModuleSet.H"
#include "Watcher.H"
#include "WatcherAux.H"
#include "History.H"
#include "SquadCombatAI.H"
#include "TerrainModule.H"
#include "TerrainAnalysis.H"
#include "Finance.H"
#include "SimpleTerrain.H"
#include "ST_ForceField.H"
#include "MineralCluster.H"
#include "Gather_accessories.H"
#include "SpecialScouting.H"
#include "SpecialWorker.H"
#include "SimpleGather.H"
#include "SpecialBaseCommander.H"
#include "SpecialDefenseCommander.H"
#include "Nets.H"
#include "NewBase.H"


/* Estructura para acceder al estado del juego */
struct ClientAIState {
	
	GameStateModule * gsm;
	
	/* 0 si modo grafico 3d desactivado, !=0 activado */
	GfxModule * gfxm;  
  
	/* true si hay que dibujar el gfx frame */
	bool just_drew;
  
	/* true si hay que salir del juego */
	bool quit;
  
	/* true si ha ocurrido un error */
	bool error;

	/* Constructor */	
	ClientAIState(GameStateModule * gsm_, GfxModule * gfxm_) :	gsm(gsm_), gfxm(gfxm_){		
		just_drew = quit = error = false;		
	}	
};


/* Clase ClientEventHandler */
class ClientEventHandler : public EventHandler {

public:
	
	/* Estructura ClienteAIState */
	ClientAIState &state;
	
	ModuleSet & ms;
  
	/* Object Scouting */
	SpecialScouting* scouting;
	
	SimpleGather* simpleGather;
		
	Vector<SpecialBaseCommander*> bases;
  
	/* Number Tiles Known */
	short nTilesKnown;
  	
	bool init;
  
	bool needGather;
  
	/* Vector de workers */
	Vector<GameObj*> workers;
	
	Vector<GameObj*> marines;
	
	Vector<GameObj*> tanks;
  
	/* Vector de Control Centers */
	Vector <GameObj *> controlCenters;
	
	static const unsigned NUMBER_OF_SCOUTS = 2;
  
	/* Constructor */
	ClientEventHandler(ClientAIState &state_, sint4 seed, ModuleSet & msGeneral, ModuleSet &msGather) : state(state_), ms(msGather), fm(msGather.get_finance()){
		rand.srand(seed);
		nTilesKnown = 0;		
		init = false;
		needGather = true;
				
		scouting = new SpecialScouting(seed, msGeneral, msGather);
		
		simpleGather = new SimpleGather(state.gsm, seed, &workers, msGather, &controlCenters);
		simpleGather->init();
		
		nets = new Nets();
		
		attack1 = NULL;
		attack2 = NULL;
		
		newBase = NULL;
		nControlCenters = 1;
		sergio = 0;
		buildCC = false;
	}
  
	
	~ClientEventHandler(){
		delete scouting;
		delete simpleGather;
		delete nets;		
	}

	/* Propotipo del metodo handle_event */
	bool handle_event(const Event &e);
    

private:


	Finance::Module & fm;
	
	Nets* nets;
	
	SpecialDefenseCommander * attack1;
	
	SpecialDefenseCommander * attack2;
	
	map<string, ScalarPoint> enemyControlCenters;	
	map<string, ScalarPoint> otherEnemyBuildings;
	map<string, ScalarPoint>::iterator it;
	
	size_t nControlCenters;
	
	Vector<sint4> posBase;
	
	bool buildCC;
		
	/* Prototipo del metodo compute_actions */
	void compute_actions();	
  
	/* Prototipo del metodo getObjects */
	void initGame(const Game::ObjCont &objs);
	
	void asignarClusters(Vector<boost::shared_ptr<MineralCluster> > clusters, const Map<GameTile> &map);
	
	bool rutaAccesible(TerrainBase::Loc locA, TerrainBase::Loc locB, const Map<GameTile> &map);
	
	TerrainBase::Loc getTile(TerrainBase::Loc point);
	
	Vector<boost::shared_ptr<MineralCluster> > getClustersNoAsignados(Vector<boost::shared_ptr<MineralCluster> > mineralClusters);
	
	float getSmallDistanceToMinerals(boost::shared_ptr<MineralCluster> cluster);
	
	float isPossible(boost::shared_ptr<MineralCluster> cluster);
	
	void defaultDefense(Vector<GameObj*> * vector);
	
	void createNewBase(boost::shared_ptr<MineralCluster> cluster);
	
	TerrainBase::Loc getLocationEnemyToNewBase(TerrainBase::Loc locCluster);
	
	void buildNewBase(const Map<GameTile> &map);
	
	void examineProblems();
	
	void evaluateAttack(SpecialDefenseCommander ** attack, bool weAreUnderAttack);
	
	void createAttack(std::string tipo);
	
	sint4 getPosBase(TerrainBase::Loc locCC);

  
	/* Objeto random */
	Random rand;
	
	NewBase * newBase;
	
	int sergio;
};

#endif
