/*
 *  SimpleGather.C
 *  
 *
 *  Created by Sergio on 11/06/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#include "SimpleGather.H"

using namespace std;

const static real8 PI = 3.141592653589793;
const static real8 DIST = 10.0;
const static real8 LS = cos( PI * 0.125 ) * DIST;
const static real8 SS = sin( PI * 0.125 ) * DIST;


void SimpleGather::init(){
	mcGather.setStatusProcedure(make_delegate(this, &SimpleGather::handleMovementStatus));
	
	newMineralsWatchID = moduleSet.get_watcher().notifyOnCreation(this, Watcher::TypeIs("mineral"), 10, false);	    
	moduleSet.get_watcher().queryUnits(Watcher::TypeIs("mineral"), newMinerals);
}


bool SimpleGather::handle_event(const Event& e){ 	
	
	if (e.get_who() == Watcher::WatcherModule::FROM){
		if (e.get_what() == Watcher::WatcherModule::CREATION){
			const Watcher::CreationEvent* ce = dynamic_cast<const Watcher::CreationEvent*>(&e);
			
			if (ce->getID() == newMineralsWatchID){				
				newMinerals = ce->getUnits();			
			}			
		}
	}
	
	if(e.get_who() == Watcher::WatcherModule::FROM ){
		
		if(e.get_what() == Watcher::WatcherModule::DEATH ){
			
			const Watcher::DeathEvent & de = reinterpret_cast<const Watcher::DeathEvent&>(e);
			const std::set<GameObj*> dUnits = de.getUnits();
			std::set<GameObj*>::const_iterator it;
			
			for(it = dUnits.begin(); it != dUnits.end(); it++){				
				onLoseUnit((*it));				
			}
		}
	}	
	
	return false;
}


/* Añade un nuevo grupo de workers para el gather */
/* Si el grupo ya existe, se rehusa */
void SimpleGather::addWorkersGather(GameObj* gob, GameObj* controlCenter, boost::shared_ptr<MineralCluster> cluster){
	
	
	bool success = false;
	SpecialGather* group;
	
	/* Se busca si ya existe el grupo, es decir,
	 mismo mineral y control center */
	FORALL(workersGather, wg){
		if((wg->controlCenter == controlCenter) && (wg->cluster == cluster)){
			success = true;
			group = &(*wg);
		}
	}
			
	Vector<SpecialWorker*> aux;
		
	if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
		moduleSet.get_watcher().notifyOnDeath(this, gob);
		SpecialWorker* sw = new SpecialWorker(gob, this, gsm);
		aux.push_back(sw);			
			
		if(success){
			FORALL(aux, w){
				group->workers.push_back(*w);
			}
			
			aux.clear();
				
		}else{
			SpecialGather sg(aux, controlCenter, cluster);	
			workersGather.push_back(sg);			
		}
	}		
}


/* Elimina un grupo entero de workers y los introduce en el vector global. Tb borra el cc del vector global */
void SimpleGather::onLoseControlCenter(unsigned index, GameObj* cc){	

	/* Liberamos todos los workers del grupo */	
	FORALL(workersGather[index].workers, w){	
		
		GameObj* gob = (*w)->get_gob();
		(*w)->clearRoute();		
		delete (*w);
				
		moduleSet.get_watcher().cancelNotifyOnDeath(this, gob);
		
		if(gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()){			
			workers->push_back(gob);			
		}		
	}
	
	workersGather[index].workers.clear();
	
	/* Eliminamos el grupo */
	if(workersGather[index].workers.size() == 0){		
		Vector<SpecialGather>::iterator it = workersGather.begin();
		it += index;
		workersGather.erase(it);		
	}
	
	/* Eliminamos el control center del vector compartido */
	Vector<GameObj*>::iterator it2;
	if((it2 = controlCenters->find(cc)) != controlCenters->end()){
		controlCenters->erase(it2);		
	}	
	administerMinerals();	
}


/* Se elimina un worker de un grupo. Se calcula de forma aleatoria */
GameObj* SimpleGather::removeWorkerGather(){		


	unsigned pos = (rand.rand_uint4() % workersGather.size());

	assert(pos < workersGather.size());
	assert(workersGather[pos].workers.size() > 0);
	
	/* Eliminamos un worker */
	SpecialWorker* w = workersGather[pos].workers.back();
	w->clearRoute();
	GameObj *gob = w->get_gob();
	delete w;

	workersGather[pos].workers.pop_back();
	
	/* Si el primer grupo esta vacio, se elimina */
	if(workersGather[pos].workers.size() == 0){		
		workersGather.pop_back();		
	}

	assert(gob);
	moduleSet.get_watcher().cancelNotifyOnDeath(this, gob);

	return gob;
}


/* Elimina un determinado worker */
void SimpleGather::removeWorkerGather(GameObj* gob){	
	
	int index = -1;
	int group = 0;	
	
	FORALL(workersGather, wg){
		
		if(index != -1)
			continue;
	
		Vector<SpecialWorker*>::iterator it = wg->workers.begin();    
		unsigned j = wg->workers.size();
	
		/* Para cada worker del vector */
		for(unsigned i = 0; i < j; i++){
		
			/* Si es el elemento buscado se elimina */
			if((*it)->get_gob() == gob){
			
				/* Eliminamos al worker */
				(*it)->clearRoute();			
				delete (*it);				
				wg->workers.erase(it);				
				index = group;
				i = j;
			}
			it++;
		}
		group++;
	}
	
	/* Si algun grupo se queda vacio, se elimina */
	
	if((index != -1) && (workersGather[index].workers.size() == 0)){		
		Vector<SpecialGather>::iterator it2 = workersGather.begin();
		it2 += index;		
		workersGather.erase(it2);		
	}	
}

void SimpleGather::onLoseUnit(GameObj * unit){	
	removeWorkerGather(unit);
}

void SimpleGather::gather(){	
	
	unsigned elementNumber = workersGather.size(); 
			
	/* Para todos los grupos */
	for(unsigned i = 0; i < elementNumber; i++){
		
		GameObj* cc = workersGather[i].controlCenter;
		
		/* Comprobamos que el control center no haya sido destruido */
		if(cc && cc->sod.in_game && *cc->sod.x >= 0 && !cc->is_dead()){
		
			unsigned elementNumber2 = workersGather[i].workers.size();
						
			/* Para cada unidad del grupo */
			for(unsigned j = 0; j < elementNumber2; j++){		
			
				/* Obtenemos un SpecialWorker */			
				SpecialWorker * worker = workersGather[i].workers[j];
			
				GameObj* gob = worker->get_gob();
			
				if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
								
					/* Si ya tiene ruta */
					if (worker->hasRoute()){

						/* Si su estado es RETURNING y esta bastante cerca del cc, deposita los minerales en el cc */
						if (worker->getState() == SpecialWorker::RETURNING){
						
							if (worker->closeEnoughToDropOff()){
								worker->dropOffMinerals();
							}			
						}        
					
						/* Si su estado es recolectando o sin trabajo asignado */
						if (worker->getState() == SpecialWorker::IDLE || worker->getState() == SpecialWorker::MINING){
						
							/* Si ha recolectado menos de 10 minerales */
							if (worker->getCurrentMinerals() < 10){
							
								/* Si el worker esta lo bastante cerca de los minerales, se le asigna la accion mine */
								if (worker->closeEnoughToMine()){
									worker->mine();
								
								/* Si el worker tiene asignado un mineral activo se le mueve hacia el*/
								}else if (worker->getMineralRec()->isActive()){
									worker->moveToMineral();
								
								/* Sino, es que el mineral ha terminado */
								}else{
									mineralDied = true;
								}
							
							/* Si ha recolectado 10 minerales */
							}else{
							
								/* Si esta lo suficientemente cerca del cc deposita los minerales */
								if (worker->closeEnoughToDropOff()){
									worker->dropOffMinerals();
								
								/* Sino, se le asigna moverse hasta el cc */
								}else{
									worker->moveToCC();
								
								}
							}
						}
					
					/* Si no tiene ruta asignada */	
					}else if(routeAllocator->getNumAvailableRoutes() > 0){					
						
						/* En cada ruta solo puede haber 1 worker */
						/* Si hay mas rutas disponibles que trabajadores */
						if (routeAllocator->getNumAvailableRoutes() > workers->size()){						
						
							/* Obtenemos la mejor ruta */
							boost::shared_ptr<SpecialRoute> bestRoute(routeAllocator->getBestRoute(worker->getLocation(), workersGather[i].controlCenter, workersGather[i].cluster));
						
							/* Si la mejor no es nula se la asignamos */
							if(bestRoute.get() != 0){								
								worker->setRoute(bestRoute);						
							}
						
						/* Sino, el mineral se ha terminado */
						}else{							
							mineralDied = true;
						}						
					
					}else{
						// Cannot use worker right now						
					}
				
					/* Si el mineral se ha terminado hay que salir */
					if (mineralDied)
						break;
				
				}
			}
		
		}else{			
			onLoseControlCenter(i, cc);
		}	
    }	
}


/* Administra los nuevos minerales y genera las rutas de acceso*/
void SimpleGather::administerMinerals(){	
    std::set<const History::Record*> mineralRecsTemp = mineralRecs;
	
    bool isNew = true;	
    
    if (!mineralDied){
        FORALL(newMinerals, nmit){
            sint4 x,y;
            (*nmit)->get_center(x,y);
            FORALL(mineralRecsTemp, mrit){
                if (*nmit == (*mrit)->getObject()){
                    isNew = false;
                    break;
                }
            }
			
            if (!isNew)
                break;
        }
    } 
	
    if (isNew){		

        std::vector<const History::Record*> rec;
        history.queryActiveRecords(Watcher::TypeIs("mineral"), rec);
        mineralRecs.clear();
        
		FORALL(rec, r){
            mineralRecs.insert(*r);
        }
		
        newMinerals.clear();        
		        
		routeAllocator = SpecialRouteAllocatorPtr(new SpecialRouteAllocator(6));
		
        computeMineralClusters();
		
		/* Se guardan las rutas para todos los control centers */
		for(unsigned i = 0; i < controlCenters->size(); i++)
			computeRoutes(i);
		
		
        if(controlCenters->empty() == false)
            routeAllocator->sortRoutesAccordingToLocation();
			
		/* Desasignamos a todos los workers sus rutas */
		for(unsigned i = 0; i < workersGather.size(); i++){
			
			/* Para cada unidad del grupo */
			for(unsigned j = 0; j < workersGather[i].workers.size(); j++){
				SpecialWorker* w = workersGather[i].workers.back();
				w->clearRoute();				
			}			
		}		
    }
    mineralDied = false;
    newMinerals.clear();	        
}


void SimpleGather::computeMineralClusters(){
	
    mineralClusters.clear();

	if (mineralRecs.empty()){
        // Do nothing		
    }else if (mineralRecs.size() <= 10){
        boost::shared_ptr<MineralCluster> cluster(new MineralCluster(mineralRecs));
        mineralClusters.push_back(cluster);
		
    }else{
        // List all mineral locations		
        Vector<sint4> clusterOf;
        Vector<ScalarPoint> mineralLocations;
        mineralLocations.reserve(mineralRecs.size());
        FORALL(mineralRecs, mit){
            const TerrainBase::Loc pos = (*mit)->getPosition();
            ScalarPoint point(pos.x, pos.y);
            mineralLocations.push_back(point);
        }
        
        // Cluster
        simple_clustering(mineralLocations, 2, 8, clusterOf);
        // Store the results
        sint4 clusterCount = *max_element(clusterOf.begin(), clusterOf.end()) + 1;
        mineralClusters.reserve(clusterCount);
        
        FORU(i, clusterCount){
            boost::shared_ptr<MineralCluster> cluster(new MineralCluster());
            mineralClusters.push_back(cluster);			
        }
		
        Vector<sint4>::iterator cit = clusterOf.begin();
        FORALL(mineralRecs, mrit){
            mineralClusters[*cit]->addMineral(*mrit);
            ++cit;
        }
    }
    
    // Compute cluster centroids
    FORALL(mineralClusters, mcit){
        (*mcit)->calculateCentroid();		
    }

	changes = true;
}


void SimpleGather::computeRoutes(int indexControlCenters){
    
    if(controlCenters->empty())
        return;
	
	/* Obtiene las coordenadas del centro del Control Center */
    TerrainBase::Loc ccLoc;
	
	Vector<GameObj*> vectorCC = *controlCenters;
	
    vectorCC[indexControlCenters]->get_center(ccLoc.x, ccLoc.y);
	
    /* cluster es un iterador de los elementos de mineralClusters */
    Vector<MineralClusterPtr>::iterator cluster = mineralClusters.begin();
    
	/* Para cada cluster de minerales */
	for (cluster = mineralClusters.begin(); cluster != mineralClusters.end(); cluster++){
        
		/* Obtiene los minerales que hay en el cluster */
		set<const History::Record*> mineralsInCluster;
        (*cluster)->getMinerals(mineralsInCluster);
		
		/* Vector que contiene para cada mineral del cluster los puntos iniciales y el centro del mineral */
        Vector<MineralPtr> minerals;
		
		/* Para cada mineral del cluster */
        FORALL(mineralsInCluster, record){
			
            /* Crea los puntos iniciales */
            TerrainBase::Loc c = (*record)->getPosition();
            Vector<TerrainBase::Loc> tempPoints;
			
            sint4 ls(static_cast<sint4>(LS));
            sint4 ss(static_cast<sint4>(SS));
            tempPoints.push_back(TerrainBase::Loc(c.x + ls, c.y + ss));
            tempPoints.push_back(TerrainBase::Loc(c.x + ss, c.y + ls));
            tempPoints.push_back(TerrainBase::Loc(c.x - ss, c.y + ls));
            tempPoints.push_back(TerrainBase::Loc(c.x - ls, c.y + ss));
            tempPoints.push_back(TerrainBase::Loc(c.x - ls, c.y - ss));
            tempPoints.push_back(TerrainBase::Loc(c.x - ss, c.y - ls));
            tempPoints.push_back(TerrainBase::Loc(c.x + ss, c.y - ls));
            tempPoints.push_back(TerrainBase::Loc(c.x + ls, c.y - ss));
            
			/* Crea un nuevo objeto Mineral y lo introduce en el vector */
            boost::shared_ptr<Mineral> mineral(new Mineral(tempPoints, c));
            minerals.push_back(mineral);
        }
		
		/* Para cada objeto mineral del vector minerals (objeto que cuenta con los puntos iniciales y el centro del mineral)*/
        FORALL(minerals, mineralToTest){                      
            
			/* Para cada punto inicial del mineral */
            FORALL_NI((*mineralToTest)->points, pointToTest){
                
                bool clashes = false;
				
				/* Comprueba si el punto pertenece al lado equivocado */
                if (pointToTest->distance((*cluster)->getCentroid()) < (*mineralToTest)->centre.distance((*cluster)->getCentroid())){                    
                    pointToTest = (*mineralToTest)->points.erase(pointToTest);
                    continue;
                }
				
				/* Para cada mineral del vector minerals, es decir, para cada mineral del cluster */
                FORALL(minerals, otherMineral){
					
					/* Si el mineral que estamos evaluando es el mismo que el del punto actual se pasa al siguiente */
					if (mineralToTest == otherMineral){
                        continue;
                    }					
					
					/* Comprueba si el punto inicial que se esta evaluando esta demasiado cerca de otro mineral */
                    if (pointToTest->distance((*otherMineral)->centre) < 9){                        
                        clashes = true;
                        pointToTest = (*mineralToTest)->points.erase(pointToTest);
                        break;
                    }
                }
				
				/* Si el punto es conflictivo se pasa al siguiente */
                if (clashes)
				    continue;
				
				/* Para cada mineral del vector minerals, es decir, para cada mineral del cluster */
                FORALL(minerals, otherMineral){                    
					
					/* Si el mineral que estamos evaluando es el mismo que el del punto actual se pasa al siguiente */
					/* Si la distancia del punto que estamos evaluando respecto al centro del otro mineral es elevada pasamos al siguiente punto */
                    if (otherMineral == mineralToTest || pointToTest->distance((*otherMineral)->centre) > 15){
                        continue;
                    }
					
					/* Para cada punto inicial del otro mineral */
                    FORALL((*otherMineral)->points, otherPoint){                        						
                        
						/* Si el punto que se esta evaluando esta demasiado cerca de algun punto inicial del otro mineral se pasa al siguiente */
						if (pointToTest->distance(*otherPoint) < 8){                            
                            clashes = true;
                            pointToTest = (*mineralToTest)->points.erase(pointToTest);
                            break;
                        }
                    }
					
                    if (clashes) break;
                }
                
				/* Si el punto es conflictivo se pasa al siguiente */
				if (clashes) continue;
				
				/* Pasamos al siguiente punto */
                ++pointToTest;
            }
        }		
        
        Vector<boost::shared_ptr<Mineral> >::iterator mit = minerals.begin();
        
		/* Para cada mineral del cluster */
		FORALL(mineralsInCluster, mrit){
			
			/* Para cada punto inicial del mineral */
			FORALL((*mit)->points, p){
				
                /* Se añade la ruta para el mineral */
				
				RoutePtr route(new SpecialRoute(this, *mrit, *p, vectorCC[indexControlCenters]));
                routeAllocator->addRoute(route);
            }
            ++mit;
        }
    }   
}


void SimpleGather::moveTo(GameObj *unit, GameObj *gob){
    mcGather.moveUnit(unit, Movement::TouchObject(gob));
}


void SimpleGather::moveTo(GameObj *unit, const TerrainBase::Loc &location){
    mcGather.moveUnit(unit, Movement::TouchPoint(location));
}


void SimpleGather::handleMovementStatus(const Movement::Task & task){
    FORALL( task.getUnits(), u ){
		const GameObj * unit(*u);		
		
		/* Para todos los grupos */
		for(unsigned i = 0; i < workersGather.size(); i++){
			
			/* Para cada unidad del grupo */
			for(unsigned j = 0; j < workersGather[i].workers.size(); j++){				
				
				/* Obtenemos un SpecialWorker */			
				SpecialWorker * worker = workersGather[i].workers[j];		
				
				if (worker->get_gob() == unit){					
					
					switch( task.getStatus()){
							
						case Movement::Task::PLAN_FAILURE:
							handlePlanFailure(worker);
							break;
							
						case Movement::Task::MOVE_FAILURE:
							handleMoveFailure(worker);
							break;
							
						case Movement::Task::SUCCESS:
							handleSuccess(worker);
							break;
							
						default:
							break;
					}
				}
			}
		}
    }
}


void SimpleGather::handleMoveFailure(SpecialWorker *worker){
	moveRandomly(worker->get_gob());
	worker->setState(SpecialWorker::IDLE);
}


void SimpleGather::handlePlanFailure(SpecialWorker *worker){   
	
	if (worker->getState() == SpecialWorker::MOVING){
		if (worker->hasRoute()){
			RoutePtr route = worker->getRoute();
			if(getDistance(worker->get_gob(), route->getAccessPoint()) == -1){                
				worker->clearRoute();
				routeAllocator->removeRoute(route);
			}
		}
	}else if (worker->getState() == SpecialWorker::RETURNING){
		moveRandomly(worker->get_gob());
	}
	worker->setState(SpecialWorker::IDLE);	
}


void SimpleGather::handleSuccess(SpecialWorker *worker){
    if(worker->hasRoute()){
        worker->getRoute()->planSuccessOccurred();
    }
    worker->setState(SpecialWorker::IDLE);
}


real8 SimpleGather::getDistance(GameObj *obj, TerrainBase::Loc loc){
    Movement::Path::ptr path(mcGather.findPath(obj, Movement::TouchPoint(loc)));
    
	if(!path)
		return -1;
	
    real8 dist = path->getLength();   
    real8 finalDist = dist; //0.0;
    return finalDist;
}


void SimpleGather::moveRandomly(GameObj *unit){
    ScalarPoint t;
    unit->get_center(t.x, t.y);
    t.x += (rand.rand_uint4() % 10) - 5;
    t.y += (rand.rand_uint4() % 10) - 5;
	
    Vector<sint4> params;
    params.push_back(t.x);
    params.push_back(t.y);
    params.push_back(*(unit->sod.max_speed));
    unit->set_action("move", params);
	
}


unsigned SimpleGather::getNumberWorkersGather(){
	return workersGather.size();
}


unsigned SimpleGather::getNumberWorkers(unsigned i){
	return workersGather[i].workers.size();
}


bool SimpleGather::getMineralDied(){
	return mineralDied;
}


bool SimpleGather::isEmptyNewMinerals(){
	return newMinerals.empty();
}


unsigned SimpleGather::getNumberMineralClusters(){
	return mineralClusters.size();
}


void SimpleGather::addControlCenter(GameObj* cc){
	controlCenters->push_back(cc);	
}


size_t SimpleGather::countWorkers(){

	size_t count = 0;

	for(size_t i = 0; i < workersGather.size(); i++)
		count += workersGather[i].workers.size();
		
	return count;
}

Vector<boost::shared_ptr<MineralCluster> > SimpleGather::getMineralClusters(){		
	return mineralClusters;	
}

bool SimpleGather::newChanges(){
	return changes;
}

void SimpleGather::setChanges(bool value){
	changes = value;
}


float SimpleGather::canGather(Vector<boost::shared_ptr<MineralCluster> > clusters, size_t workersQueued){
	float success = 0.0;
	size_t count = 0;
	
		
	/* Se buscan todos los clusters de la base */
	FORALL(clusters, c){
	
		/*  Se busca en todos los grupos */
		FORALL(workersGather, wg){
			if(wg->cluster == *c)
				count += wg->workers.size();			
		}	
	}
			
	count += workersQueued;
		
	if((10 * clusters.size()) > count)
		success = 1.0;	

	return success;
}



boost::shared_ptr<MineralCluster> SimpleGather::getBestCluster(Vector<boost::shared_ptr<MineralCluster> > clusters){

	boost::shared_ptr<MineralCluster> cluster;
	size_t min = 11;

	/* Se buscan todos los clusters de la base */
	FORALL(clusters, c){
		
		bool success = false;

		/*  Se busca en todos los grupos */
		FORALL(workersGather, wg){

			/* Si el cluster del grupo actual es el mismo que el cluster actual de la base */
			if(wg->cluster == *c){
			
				success = true;

				if((min == 11) || (wg->workers.size() < min)){
					min = wg->workers.size();
					cluster = wg->cluster;
				}
			}
		}
		
		/* Si no se ha encontrado es que no se ha creado el grupo, por lo que tiene 0 workers asignados */
		if(success == false){
			
			if(0 < min){
				min = 0;
				cluster = *c;
			}
		}
	}
		
	return cluster;
}



void SimpleGather::comprobarWorkersGather(){

	bool success;
	int count = 0;
			
	/* Examinamos todos los grupos */
	FORALL(workersGather, wg){
	
		success = false;
	
		/* Buscamos el cluster del grupo en el vector de clusters */
		FORALL(mineralClusters, mc){
			
			/* Si se encuentra se marca */
			if(wg->cluster == *mc)
				success = true;			
		}
		
		/* Si no se ha encontrado el cluster, se liberan todas las unidades */
		if(success == false){			
			
			/* Se contabilizan los grupos que hay que borrar */
			count++;
			
			FORALL(wg->workers, w){
				
				//SpecialWorker* sw = (*w)->back();
				(*w)->clearRoute();
				GameObj *gob = (*w)->get_gob();
				delete (*w);
				//w->pop_back();
					
				moduleSet.get_watcher().cancelNotifyOnDeath(this, gob);
				
				if(gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()){					
					workers->push_back(gob);
				}
			}
			wg->workers.clear();
		}
	}

	/* Eliminamos todos los grupos vacios */
	for(int i = 0; i < count; i++){

		size_t aux = workersGather.size();
		Vector<SpecialGather>::iterator it = workersGather.begin();

		/* Se busca el primer grupo vacio */
		for(size_t j = 0; j < aux; j++){

			if(it->workers.size() == 0){
				workersGather.erase(it);
				j = aux;
			}		
			it++;
		}
	}
}