/*
 *  SimpleGather.H
 *  
 *
 *  Created by Sergio on 11/06/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef SimpleGather_H
#define SimpleGather_H


#include "Random.H"
#include "Game.H"
#include "Movement.H"
#include "ModuleSet.H"
#include "Movement.H"
#include "MineralCluster.H"
#include "Gather_accessories.H"
#include "Delegate.H"
#include "SpecialWorker.H"
#include "SpecialRouteAllocator.H"
#include "History.H"
#include "Watcher.H"
#include "WatcherAux.H"


class SpecialWorker;
class GameStateModule;
class GfxModule;


typedef boost::shared_ptr<SpecialRouteAllocator> SpecialRouteAllocatorPtr;
typedef boost::shared_ptr<MineralCluster> MineralClusterPtr;


class Mineral {
public:
    Vector<TerrainBase::Loc> points;
    TerrainBase::Loc centre;
    Mineral(Vector<TerrainBase::Loc> points_, TerrainBase::Loc centre_)
	: points(points_), centre(centre_)
    {}
    ~Mineral(){}
};


typedef boost::shared_ptr<Mineral> MineralPtr;


struct SpecialGather{
	
	/* workers */
	Vector<SpecialWorker*> workers;
	
	/* ControlCenter where deposit the minerals */
	GameObj* controlCenter;
	
	/* Cluster of Mineral where the workers must mining */	
	boost::shared_ptr<MineralCluster> cluster;
	
	SpecialGather(Vector<SpecialWorker*> workers_, GameObj* controlCenter_, boost::shared_ptr<MineralCluster> cluster_): workers(workers_){
		controlCenter = controlCenter_;
		cluster = cluster_;		
	}	
};


class SimpleGather: public EventHandler{

public:

	Vector<GameObj*>* controlCenters;
		 
	/* Propotipo del metodo handle_event */
	bool handle_event(const Event &e);
	
	void moveTo(GameObj *unit, GameObj *gob);	
	
	void moveTo(GameObj *unit, const TerrainBase::Loc &location);
		
	void gather();	
	
	//void addWorkersGather(Vector<GameObj*> elements, GameObj* controlCenter, int indexMineralCluster);
	void addWorkersGather(GameObj* gob, GameObj* controlCenter, boost::shared_ptr<MineralCluster> cluster);
	
	GameObj* removeWorkerGather();
		
	unsigned getNumberWorkersGather();
	
	unsigned getNumberWorkers(unsigned i);
	
	bool getMineralDied();
	
	bool isEmptyNewMinerals();
	
	void administerMinerals();
	
	unsigned getNumberMineralClusters();
	
	void addControlCenter(GameObj* cc);
	
	void init();
	
	size_t countWorkers();
	
	Vector<boost::shared_ptr<MineralCluster> > getMineralClusters();
	
	bool newChanges();	
	
	void setChanges(bool value);
	
	float canGather(Vector<boost::shared_ptr<MineralCluster> > clusters, size_t workersQueued);
	
	boost::shared_ptr<MineralCluster> getBestCluster(Vector<boost::shared_ptr<MineralCluster> > clusters);
	
	void comprobarWorkersGather();
	
	void computeRoutes(int indexControlCenters);
	
	void onLoseUnit(GameObj * unit);
	
	SimpleGather(GameStateModule * gsm_, sint4 seed, Vector<GameObj*>* vector, ModuleSet & msGather, Vector<GameObj*>* vectorCC): 
	 moduleSet(msGather), gsm(gsm_), mcGather(msGather.get_movement(), "Default", "Default"), history(msGather.get_history()){
		rand.srand(seed);		
		mineralDied = false;
		routeAllocator = SpecialRouteAllocatorPtr(new SpecialRouteAllocator());
		workers = vector;
		controlCenters = vectorCC;
		msGather.get_history().trackBlueprint("mineral");
		changes = false;
	}
	
	~SimpleGather(){
		moduleSet.get_watcher().cancelNotify(newMineralsWatchID);
	}

private:

	uint4 newMineralsWatchID;
	
	ModuleSet& moduleSet;
	
	GameStateModule * gsm;
	
	Movement::Context mcGather;
	
	Vector<SpecialGather> workersGather;
	
	History::HistoryModule &history;

	//Vector<GameObj*> controlCenters;
	
	Vector<GameObj*>* workers;

	// Contains any minerals that came into view (or out of fog of war) since the last frame.
	std::set<GameObj*> newMinerals;
	
	bool mineralDied;
	
	// a vector of mineral clusters containing organised Minerals
	Vector<boost::shared_ptr<MineralCluster> > mineralClusters;
	
	// Contains the History records.  All calls to minerals are done through these
	// records (for safety in fog of war).
	std::set<const History::Record*> mineralRecs;
	
	// a class that sorts and allocates the routes
	boost::shared_ptr<SpecialRouteAllocator> routeAllocator;
	
	Random rand;
	
	bool changes;
	
	void computeMineralClusters();
	
	//void computeRoutes(int indexControlCenters);
	
	/**
     * called whenever a status update is received
     * @param task the task object whose status has changed
     */
    void handleMovementStatus(const Movement::Task & task);
    
    /**
     * called whenever a unit cannot move to its next point
     * @param worker the worker that cannot move
     */
    void handleMoveFailure(SpecialWorker *worker);
	
    /**
     * called whenever a unit cannot find a path to its destination
     * @param worker the worker that cannot find a path
     */
    void handlePlanFailure(SpecialWorker *worker);
	
    /**
     * called whenever a unit reaches its destination.
     * @param worker the worker that reached its destination
     */
    void handleSuccess(SpecialWorker *worker);
	
	real8 getDistance(GameObj *obj, TerrainBase::Loc loc);
	
	void moveRandomly(GameObj *unit);
	
	void onLoseControlCenter(unsigned index, GameObj* cc);
	
	void removeWorkerGather(GameObj* gob);	
};

#endif
