#ifndef SpecialBuildCommander_H
#define SpecialBuildCommander_H

#include "SpecialCommander.H"
#include "ClientModule.H"
#include "TerrainBase.H"
#include "Watcher.H"
#include "Finance.H"
#include "Movement.H"

class SpecialBuildCommander : public SpecialCommander, ClientModule{

    enum STATE { PLANNING, ENROUTE, READY, BUILDING, STATE_FINISHED, STATE_FAILED };

    TerrainAnalysis::Module& tam;
    Watcher::WatcherModule&  wm;
    Movement::Context        mc;

    std::string            bp_name;
	TerrainBase::Loc       nearLoc;

    Finance::Ticket *      funds;

    GameObj *              worker;
    GameObj *              building;

    STATE                  state;

    Vector<TerrainBase::Loc> buildSites;
    size_t		             nextSite;

    bool                   tryingToBuild;
    sint4                  countDown;
    Watcher::WatchID       buildWatch;
    Watcher::WatchID       completeWatch;

    void doPlanning();
    void tryToBuild();
    bool findBuildSite();
    void moveToNextSite();

    void onFail();
	
public:

    SpecialBuildCommander(SpecialCommander*  mySuperior, 
						  GameObj*            worker,
						  const std::string & bp_name,
						  TerrainBase::Loc    nearLoc,
						  Finance::Ticket *   funds = 0,
						  sint4               updateTime_ = 25);
						  
    ~SpecialBuildCommander();

    void onReceiveUnit(GameObj * unit, Watcher::WatchID watch);
    void onLoseUnit(GameObj * unit);
    bool onEvent(const Event& event);
    void handleMovementStatus(const Movement::Task & task);

    void exec();

    const std::string & getBuildingName() const { return bp_name; }

    static const sint4 FROM; 
    static const sint4 FINISHED;
    static const sint4 FAILED;
};

class BuildFinishedEvent : public Event{

    SpecialBuildCommander * source;
    GameObj *        worker;
    GameObj *        building;
	
public:

    BuildFinishedEvent(SpecialBuildCommander * source, GameObj * worker, GameObj * building) :
        Event(SpecialBuildCommander::FROM, SpecialBuildCommander::FINISHED), source(source), worker(worker), building(building) {}

    ~BuildFinishedEvent() {}

    SpecialBuildCommander * getSource() const { return source; }
    GameObj *        getWorker() const { return worker; }
    GameObj *        getBuilding() const { return building; }
};

class BuildFailedEvent : public Event{

public:

    enum REASON { NO_BUILD_SITE, WORKER_KILLED, BUILDING_DESTROYED };
	
private:

    SpecialBuildCommander * source;
    GameObj *        worker;
    REASON           reason;
	
public:

    BuildFailedEvent(SpecialBuildCommander * source, GameObj * worker, REASON reason) :
        Event(SpecialBuildCommander::FROM, SpecialBuildCommander::FAILED), source(source), worker(worker), reason(reason) {}

    ~BuildFailedEvent() {}

    SpecialBuildCommander * getSource() const { return source; }
    GameObj *        getWorker() const { return worker; }
    REASON           getReason() const { return reason; }
};

#endif
