#include "SpecialDefenseCommander.H"
#include "GameStateModule.H"
#include "SquadCombatAI.H"
#include "WatcherAux.H"
#include "Delegate.H"


SpecialDefenseCommander::SpecialDefenseCommander(SpecialCommander * mySuperior, const TerrainBase::Loc & center, sint4 radius, sint4 valueLastAttack, bool valueUnderAttack) :
    SpecialCommander(mySuperior, 8),
    center(center),
    radius(radius),
    me(moduleSet.get_game_state().get_game().get_client_player()),
    underControl(false),
    mc(moduleSet.get_movement(), "Default", "Default"),
    ga(moduleSet, center.x, center.y, radius)
{	
    moduleSet.get_combat().add_handler(this);
    mc.setStatusProcedure( make_delegate(this, &SpecialDefenseCommander::onMovement) );
    
    ga.set_passable_height( ga.sample_height(center) );
    ga.set_object_border( "mineral", 24 );
    ga.set_object_border( "controlCenter", 32 );
	lastAttack = valueLastAttack;
	underAttack = valueUnderAttack;
}


SpecialDefenseCommander::SpecialDefenseCommander(ModuleSet & ms, const TerrainBase::Loc & center, sint4 radius, sint4 valueLastAttack, bool valueUnderAttack) :
	SpecialCommander(ms, 8),
	center(center),
	radius(radius),
	me(moduleSet.get_game_state().get_game().get_client_player()),
	underControl(false),
	mc(moduleSet.get_movement(), "Default", "Default"),
	ga(moduleSet, center.x, center.y, radius)
{	
    moduleSet.get_combat().add_handler(this);
    mc.setStatusProcedure( make_delegate(this, &SpecialDefenseCommander::onMovement) );
    
    ga.set_passable_height( ga.sample_height(center) );
    ga.set_object_border( "mineral", 24 );
    ga.set_object_border( "controlCenter", 32 );
	lastAttack = valueLastAttack;
	underAttack = valueUnderAttack;
}


SpecialDefenseCommander::~SpecialDefenseCommander(){
    moduleSet.get_combat().remove_handler(this);
}

GameObj * SpecialDefenseCommander::pickClosest(const Watcher::UnitPredPtr & pred, const std::set<GameObj*> & candidates) const{
    GameObj * target = NULL;
    real8 targetDistance = 0.0;

    // Determine the closest enemy
    FORALL(candidates, e){

        GameObj * enemy = (*e);
        if(pred->evaluate(enemy) == false){
            continue;
        }

        TerrainBase::Loc enemyLoc;
        enemy->get_center(enemyLoc.x, enemyLoc.y);
        real8 enemyDistance = center.distance(enemyLoc);

        if(target == NULL || enemyDistance < targetDistance){
            target = enemy;
            targetDistance = enemyDistance;
        }
    }

    return target;
}

GameObj * SpecialDefenseCommander::pickBestTarget(const std::set<GameObj*> & candidates) const{
    GameObj * target = pickClosest(Watcher::IsCombat(), candidates);
    if(target) return target;
    target = pickClosest(Watcher::IsWorker(), candidates);
    if(target) return target;
    return pickClosest(Watcher::IsBuilding(), candidates);
}

void SpecialDefenseCommander::exec(){
    
	//if(unit2state.empty()){
        // Can't do anything without units
        // TODO: Appropriate message to commander
        //underControl = false;
        //return;
    //}

    std::set<GameObj*> enemies;
    moduleSet.get_watcher().queryUnits(And(InCircle(center.x, center.y, radius), Not(OwnerIs(me)), Not(OwnerIs(2))), enemies);


    Vector<GameObj::WPtr> activeUnits; // Units which are attacking
    Vector<GameObj::WPtr> passiveUnits; // Units which are moving/idle
	
	size_t nUnits = unit2state.size();
	size_t unitsOutOfBoundary = 0;

    SquadCombatAI & scai = moduleSet.get_combat();

    // If all enemies have been defeated
    if(enemies.empty()){

        underControl = true;
		underAttack = false;

        // Iterate through all of our units
		FORALL(unit2state, u2s){
			GameObj * unit = u2s->first;
			UNITSTATE & state = u2s->second;

			scai.disable_unit_motion(unit);

			if(state == CHASING || state == ATTACKING){
				state = STOPPED;
			}

			if(state == RETURNING){
				underControl = false;
				unitsOutOfBoundary++;
			}

			if(state == STOPPED){
				sint4 boundary = (radius * 3) / 4;

				// If the unit is outside of his zone
				TerrainBase::Loc objLoc;
				unit->get_center(objLoc.x, objLoc.y);
				sint4 distance = static_cast<sint4>(center.distance(objLoc));
				
				if (distance > boundary){
					// If we can find a free spot, move to it
					TerrainBase::Loc target;
					if(ga.find_free_spot(center, boundary, target)){
						mc.moveUnit(unit, Movement::TouchPoint(target));
						state = RETURNING;
						underControl = false;
						unitsOutOfBoundary++;
					}
				
				}else{
					// Otherwise we couldn't find a free spot
					// Just wait around, we'll find one eventually
				}
			}
		}
		
		
		if((nUnits != 0) && (((float)unitsOutOfBoundary/nUnits) <= 0.25)){			
			underControl = true;
		}
    
	}else{ // There are enemies		

        GameObj * target = pickBestTarget(enemies);

		assert(target != NULL); // There should be at least one enemy

        Vector<GameObj::WPtr> wvec;

        FORALL( unit2state, u2s ){
		
			GameObj * unit = u2s->first;
            UNITSTATE & state = u2s->second;

            real8 distance = unit->distance_to(*target);
            
            if(state == STOPPED || state == ATTACKING){
                if(distance > 120.0f){
					mc.moveUnit(unit, Movement::AttackObject(target));
                    state = CHASING;
                
				}else{
                    state = ATTACKING;
                }
            
			}else{ // We are chasing or moving            
                if(distance < 80.0f){
                    state = ATTACKING;
                }
            }


            if(state == ATTACKING){
                scai.enable_unit_motion( unit );
                wvec.push_back( unit );
            
			}else{
                scai.disable_unit_motion( unit );
            }
        }

        scai.handle_event(SquadCombatEvent(commanderID, SquadCombatAI::ATTACK_TARGET, wvec, target));
        underControl = false;
		underAttack = true;
		lastAttack = moduleSet.get_game_state().get_game().get_view_frame();
    }
}

void SpecialDefenseCommander::onReceiveUnit(GameObj * unit, Watcher::WatchID){
    
	if (unit && unit->sod.in_game && *unit->sod.x >= 0 && !unit->is_dead()) {
		moduleSet.get_watcher().notifyOnDeath(this, unit);
		unit2state.insert( std::make_pair(unit, STOPPED) );

		SquadCombatAI & scai = moduleSet.get_combat();
		scai.disable_unit_motion(unit);
	}
}

void SpecialDefenseCommander::onLoseUnit(GameObj * unit){
    unit2state.erase( unit );
}

bool SpecialDefenseCommander::onEvent(const Event &){
    return false;
}

void SpecialDefenseCommander::onMovement(const Movement::Task & task){
    
	FORALL(task.getUnits(), u){
        std::map<GameObj*, UNITSTATE>::iterator it;
        it = unit2state.find(const_cast<GameObj*>(*u)); // TODO: Fix
        
		if(it != unit2state.end()){			
			it->second = STOPPED;
		}
    }
}

bool SpecialDefenseCommander::isUnderControl() const{
    return underControl;
}

Vector<GameObj*> SpecialDefenseCommander::getAllUnits(){

	Vector<GameObj*> vector;
	SquadCombatAI & scai = moduleSet.get_combat();
	
	FORALL(unit2state, it){

		GameObj* gob = it->first;
		moduleSet.get_watcher().cancelNotifyOnDeath(this, gob);
		scai.enable_unit_motion(gob);
		vector.push_back(gob);
    }
	
	unit2state.clear();
	
	return vector;
}

GameObj* SpecialDefenseCommander::getUnit(){

	GameObj* gob = NULL;
	
	/* Si al menos hay una unidad, la devolvemos y eliminamos del mapa */
	if(unit2state.size() > 0){
				
		SquadCombatAI & scai = moduleSet.get_combat();
		std::map<GameObj*, UNITSTATE>::iterator it = unit2state.begin();
		gob = it->first;
		moduleSet.get_watcher().cancelNotifyOnDeath(this, gob);
		unit2state.erase(gob);
		scai.enable_unit_motion(gob);		
	}	

	return gob;
}

size_t SpecialDefenseCommander::countUnits(const std::string & bp_name){
	
	size_t count = 0;
	
	if(bp_name == ""){
		count = unit2state.size();
		
	}else{
	
		FORALL(unit2state, u2s){
			GameObj * unit = u2s->first;
			if(unit->bp_name() == bp_name){
				count++;
			}
		}
	}
	
	return count;
}

sint4 SpecialDefenseCommander::getLastAttack(){
	return lastAttack;
}


bool SpecialDefenseCommander::isUnderAttack() const{
    return underAttack;
}


Vector<GameObj*> SpecialDefenseCommander::getUnitsToNewBase(float a){
	
	SquadCombatAI & scai = moduleSet.get_combat();
	size_t nMarines = countUnits("marine");
	size_t nTanks = countUnits("tank");
	Vector<GameObj*> vector;
		
	nMarines = (nMarines * a)/3;
	nTanks = (nTanks * a)/3;
		
	size_t j = (nMarines + nTanks);
	
	/* Se obtienen todas las unidades */
	for(size_t i = 0; i < j; i++){		
		
		size_t nUnits = unit2state.size();
		std::map<GameObj*, UNITSTATE>::iterator it = unit2state.begin();		
		
		for(size_t k = 0; k < nUnits; k++){
			GameObj* gob = it->first;
			
			if((gob->bp_name() == "marine") && (nMarines > 0)){				
				
				moduleSet.get_watcher().cancelNotifyOnDeath(this, gob);
				unit2state.erase(gob);
				scai.enable_unit_motion(gob);
				vector.push_back(gob);				
				nMarines--;
				k = nUnits;				
			
			}else if((gob->bp_name() == "tank") && (nTanks > 0)){								
				
				moduleSet.get_watcher().cancelNotifyOnDeath(this, gob);
				unit2state.erase(gob);
				scai.enable_unit_motion(gob);
				vector.push_back(gob);				
				nTanks--;
				k = nUnits;				
			}			
			
			if(k != nUnits)
				it++;
		}	
	}
	
	return vector;
}
	



