#ifndef __SPECIAL_DEFENSE_COMMANDER_H__
#define __SPECIAL_DEFENSE_COMMANDER_H__

#include "SpecialCommander.H"
#include "Game.H"
#include "Watcher.H"
#include "WatcherAux.H"
#include "ClientModule.H"
#include "TerrainBase.H"
#include "GridAbstraction.H"
#include "Movement.H"

using namespace Watcher;

class SpecialDefenseCommander : public SpecialCommander{

public:
    
	enum UNITSTATE { STOPPED,     // Unit is not moving
                     CHASING,     // Unit is pathing towards a target
                     ATTACKING,   // Unit is under control of shooter
                     RETURNING }; // Unit is pathing towards the zone
private:
    
	std::map<GameObj*, UNITSTATE> unit2state;

    TerrainBase::Loc center;
    sint4 radius;
    sint4 me;
    bool underControl;

    Movement::Context mc;
    
    TerrainAnalysis::GridAbstraction ga;
	
	sint4 lastAttack;
	
	bool underAttack;
		
public:	

    SpecialDefenseCommander(SpecialCommander * mySuperior, const TerrainBase::Loc & center, sint4 radius, sint4 valueLastAttack, bool valueUnderAttack);
	SpecialDefenseCommander(ModuleSet & ms, const TerrainBase::Loc & center, sint4 radius, sint4 valueLastAttack, bool valueUnderAttack);
    ~SpecialDefenseCommander();
	
	//void setNewValues(TerrainBase::Loc & newCenter, sint4 newRadius);

    GameObj * pickClosest(const Watcher::UnitPredPtr & pred, const std::set<GameObj*> & candidates) const;
        
    GameObj * pickBestTarget(const std::set<GameObj*> & candidates) const;

    void exec(); 
                 
    void onReceiveUnit(GameObj * unit, Watcher::WatchID);

    void onLoseUnit(GameObj * unit);

    bool onEvent(const Event & event);

    void onMovement(const Movement::Task & task);
	
    
    /// Return true if all units are within the zone and no enemies are
    /// detected inside the boundaries of the zone.
    bool isUnderControl() const;
	
	Vector<GameObj*> getAllUnits();
	
	GameObj* getUnit();
	
	size_t countUnits(const std::string & bp_name);
	
	sint4 getLastAttack();
	
	bool isUnderAttack() const;
	
	Vector<GameObj*> getUnitsToNewBase(float a);
	
};

#endif
