#include "SpecialRouteAllocator.H"

#include "SpecialRoute.H"
#include <algorithm>

#define TEST(x)

using namespace std;

///////////////////////////////CONSTRUCTORS/DESTRUCTORS/////////////////////////////

SpecialRouteAllocator::SpecialRouteAllocator() : workers(0), pos(0), idPos(0){ }

SpecialRouteAllocator::SpecialRouteAllocator(sint4 workers_) : workers(workers_), pos(0), idPos(0){ }

SpecialRouteAllocator::~SpecialRouteAllocator(){ }

////////////////////////////////////////PRIVATE/////////////////////////////////////

sint4 SpecialRouteAllocator::newID() {
    idPos++;
    return idPos;
}

/////////////////////////////////////////PUBLIC/////////////////////////////////////

void SpecialRouteAllocator::addRoute(RoutePtr route_){
        route_->setID(newID());
        routes.push_back(route_);
}

void SpecialRouteAllocator::addWorker(){
    workers++;
}

RoutePtr SpecialRouteAllocator::getBestRoute(TerrainBase::Loc locationOfWorker, GameObj* cc, boost::shared_ptr<MineralCluster> cluster){

   
	RoutePtr smallestRoute;
    real8 smallestDist = 10000;


    FORALL(routes, rit){

		if ((*rit)->getCurrentWorkers() >= (*rit)->getMaxWorkers()){			
            continue;
        }

		if(!((cc == (*rit)->getCC()) &&  (esMineralCluster((*rit)->getMineralRec(), cluster)))){			
			continue;
		}		

        real8 distance = (*rit)->getAccessPoint().distance(locationOfWorker);
        if (distance < smallestDist){
	
            smallestDist = distance;
            smallestRoute = *rit;
        }		
    }
	return smallestRoute;
} 

bool SpecialRouteAllocator::esMineralCluster(const History::Record* mineral, boost::shared_ptr<MineralCluster> cluster){
	bool esMC = false;
	
	std::set<const History::Record *> myMinerals;
	cluster->getMinerals(myMinerals);

	std::set<const History::Record *>::iterator it;
	
	for (it = myMinerals.begin(); it != myMinerals.end(); it++){
		
		if(esMC)
			break;
						
		if((*it) == mineral){
			esMC = true;			
		}	
	}
	return esMC;
}


sint4 SpecialRouteAllocator::getNumRoutes(){
    return routes.size();
}

uint4 SpecialRouteAllocator::getNumAvailableRoutes(){
    sint4 count = 0;
    FORALL(routes, rit){
        if ((*rit)->getCurrentWorkers() < (*rit)->getMaxWorkers()){
            count++;
        }
    }
    return count;
}

void SpecialRouteAllocator::markRouteAsBad(RoutePtr route){
    route->planFailureOccurred();
    if(route->getPlanFailures() > 10){
        FORALL_DYN(routes, rit){
            if (route == *rit){
                routes.erase(rit);
                break;
            }
        }
    }
}

RoutePtr SpecialRouteAllocator::getWorstUsedRoute(){
    RoutePtr worstRoute;
    Vector<RoutePtr>::reverse_iterator rit;
    for(rit = routes.rbegin(); rit != routes.rend(); rit++){
        if ((*rit)->getCurrentWorkers() > 0){
            worstRoute = *rit;
        }
    }
    return worstRoute;
}

void SpecialRouteAllocator::removeRoute(RoutePtr route_){
    FORALL_DYN(routes, rit){
        if (*rit == route_){
            routes.erase(rit);
            break;
        }
    }
}

void SpecialRouteAllocator::removeWorker(){
    workers++;
}

void SpecialRouteAllocator::removeInaccessibleRoutes(){
    Vector<RoutePtr>::iterator rit = routes.begin();
    while (rit != routes.end()){
        if ((*rit)->getRouteLength() < 1){            
            rit = routes.erase(rit);
        }else{
            rit++;
        }
    }
}

void SpecialRouteAllocator::setWorkers(sint4 numWorkers){
    workers = numWorkers;
}

void SpecialRouteAllocator::sortRoutesAccordingToLocation(){
    removeInaccessibleRoutes();
    sort(routes.begin(), routes.end(), CompareRoutesByDistanceToLocation());
}

size_t SpecialRouteAllocator::countWorkersNeeded() const{
    size_t count = 0;
    FORALL(routes, rit){
        count += (*rit)->getMaxWorkers();
    }
    return count;
}