/*
 *  SpecialScouting.C
 *  
 *
 *  Created by Sergio on 09/06/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#include "SpecialScouting.H" 

using namespace std;

// tile types from core.bp
// fixme: game script reflection would help here
// in the competition we won't use water tiles
// TILE_CLIFF: non-traversable by ground units

enum { TILE_UNKNOWN=0, TILE_WATER=1, TILE_GROUND=2, TILE_CLIFF=3 };


/**
 * Añadimos un nuevo scout
 */ 
void SpecialScouting::addScout(GameObj* gob){
			
	/* Si el objeto existe, esta en juego y dentro del mapa (en el servidor), y no esta muerto */
	if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {
		moduleSet.get_watcher().notifyOnDeath(this, gob);
		Scouting workerScout(gob);
		scouts.push_back(workerScout);		
	}	
}


/**
 * Eliminamos el scout que menos tiempo lleva realizando scouting
 */
GameObj* SpecialScouting::removeScout(){
		
	GameObj * gob = scouts.back().scout;					
	scouts.pop_back();
	
	moduleSet.get_watcher().cancelNotifyOnDeath(this, gob);
	return gob;	
}


/**
 * Eliminamos un determinado scout
 */
void SpecialScouting::removeScout(GameObj* gob){

	Vector<Scouting>::iterator it = scouts.begin();
	size_t j = scouts.size();

	/* Se recorre todo el vector hasta encontrar a la unidad */
	for(size_t i = 0; i < j; i++){

		/* Si es el elemento buscado se elimina */
		if(it->scout == gob){
			scouts.erase(it);
			i = j;			
		}
		it++;
	}
}


/**
 * Devuelve el número de objetos que estan realizando scouting.
 */
unsigned SpecialScouting::getNumberScouts(){	
	return scouts.size();
}


/**
 * Realiza scouting con los objetos del vector scouts.
 */
void SpecialScouting::scouting(const Map<GameTile> &map){

	/* Para cada scaut hay que comprobar su estado, que puede ser:
	 Sin destino, en destino, parado, en movimiento, pillado o caso no identificado */
	for(unsigned i = 0; i < scouts.size(); i++){	
		
		/* is scouting not started */
		if(!scouts[i].running){	
			
			/* Obtenemos las coordenadas del objetivo y las asignamos */
			ScalarPoint aux = getTargetTile(map);
			scouts[i].target.x = aux.x;
			scouts[i].target.y = aux.y;
			
			TerrainBase::Loc target(aux.x, aux.y);									
			mc.moveUnit(scouts[i].scout, Movement::TouchPoint(target));
			scouts[i].running = true;
			scouts[i].contador = 0;					
			
		/* En destino */				
		}else if(getDistanceToTarget(scouts[i].target, *scouts[i].scout->sod.x, *scouts[i].scout->sod.y) < 48){
			scouts[i].running = false;					
			
		/* Esta parado */
		}else if((*scouts[i].scout->sod.speed == 0) && (scouts[i].contador < 50)){
			scouts[i].contador++;
			
		/* Esta en movimiento */
		}else if(*scouts[i].scout->sod.speed != 0){
			scouts[i].contador = 0;
			
		/* Estoy pillado */
		}else if(scouts[i].contador >= 50){					
			scouts[i].running = false;			
		}		
	}	
}



/**
 * Obtenemos un area desconocida.
 * @return ScalarPoint contiene la coordenada del centro del tile destino.
 */
ScalarPoint SpecialScouting::getTargetTile(const Map<GameTile> &map){
	
	/* Calculamos un area desconocoda */	
	bool target = false;
	ScalarPoint point;
	
	while(!target){
		point.x = rand.rand_uint4() % map.get_width();
		point.y = rand.rand_uint4() % map.get_height();
		const GameTile &tile = map(point.x, point.y);
		
		/* Tile objetivo desconocida */
		if(!(tile.is_split()) && (tile.get_typeW() == TILE_UNKNOWN)){
			
			/* Si tile objetivo es limite izquierdo */
			if(point.x == 0){				
				const GameTile &tileAux2 = map(point.x+1, point.y);
				if((!tileAux2.is_split()) && (tileAux2.get_typeW() == TILE_UNKNOWN))
					target = true;
				
				/* Si tile objetivo es limite dereche */		
			}else if(point.x == map.get_width()-1){
				const GameTile &tileAux1 = map(point.x-1, point.y);
				if((!tileAux1.is_split()) && (tileAux1.get_typeW() == TILE_UNKNOWN))
					target = true;
				
				/* Si tile objetivo no es limite */				
			}else{
				const GameTile &tileAux1 = map(point.x-1, point.y);
				const GameTile &tileAux2 = map(point.x+1, point.y);				
				if((!tileAux1.is_split()) && (tileAux1.get_typeW() == TILE_UNKNOWN) 
				   && (!tileAux2.is_split()) && (tileAux2.get_typeW() == TILE_UNKNOWN))
					target = true;
			}						
		}		
	}	
		
	point.y = (point.y * 16) + 8;
	point.x = (point.x * 16) + 8;	
	
	return point;
}


/*
 * Calcula la distancia del punto targetScaut al formado por las coordenadas x e y.
 */
real8 SpecialScouting::getDistanceToTarget(ScalarPoint targetScaut, sint4 x, sint4 y){
	real8 distancia = square(real8(targetScaut.x - x)) + square(real8(targetScaut.y - y));
	distancia = sqrt(distancia);	
	return distancia;	
}


void SpecialScouting::onLoseUnit(GameObj * unit){	
	removeScout(unit);
}


bool SpecialScouting::handle_event(const Event &event){
			
	if( event.get_who() == Watcher::WatcherModule::FROM ){
			
		if( event.get_what() == Watcher::WatcherModule::DEATH ){
				
			const Watcher::DeathEvent & e = reinterpret_cast<const Watcher::DeathEvent&>(event);
			const std::set<GameObj*> dUnits = e.getUnits();
			std::set<GameObj*>::const_iterator it;
				
			for(it = dUnits.begin(); it != dUnits.end(); it++){				
				onLoseUnit((*it));				
			}
		}
	}	
	return false;
}

