#include "SpecialWorker.H"

#include "SpecialRoute.H"
#include "PlayerInfo.H"
#include "GameStateModule.H"
#include "Game.H"
#include "GameObj.H"
#include "ClientEventHandler.H";

const std::string miningComponent = "pickaxe";
const std::string miningAction = "mine";
const std::string returnResourcesAction= "return_resources";

/////////////////////////////// Constructors ///////////////////////////
SpecialWorker::SpecialWorker(GameObj* gob_,SimpleGather *commander_,GameStateModule *gsm_)
    : gsm(gsm_),
      commander(commander_),	  
      //gob(gob_),
	  
      state(IDLE),
      knowsRoute(false)
{

gob = gob_;
}

SpecialWorker::~SpecialWorker(){ 
    if(route != 0){
        route->removeWorker();
    }
}

//////////////////////////////////////PUBLIC///////////////////////////////////////

void SpecialWorker::clearRoute(){
    if(route != 0){
        route->removeWorker();
        route.reset();
    }	
}


bool SpecialWorker::closeEnoughToDropOff() const{
    assert(route != 0);
    sint4 dist = (sint4)(Game::distance(*gob, *route->getCC()));
    return dist <= WORKER_CC_DISTANCE_THRESHOLD;
}


bool SpecialWorker::closeEnoughToMine() const {
    assert(route != 0);

    if (route->mineralIsActive()){
        sint4 dist = (sint4)(Game::distance(*gob, *(route->getMineralRec()->getObject())));
        if (dist > WORKER_MINERAL_DISTANCE_THRESHOLD){
            commander->moveTo(gob, route->getMineralRec()->getObject());
            return false;
        }
        else
            return true;
    }
    return false;
}


void SpecialWorker::dropOffMinerals(){
    assert(route != 0);

    PlayerInfo& pi = gsm->get_game().get_cplayer_info();
    Vector<sint4> args(1, pi.get_id(route->getCC()));
    gob->set_action(returnResourcesAction, args);
    setState(IDLE);
}


GameObj* SpecialWorker::getControlCenter() const{
    assert(route != 0);
    return route->getCC();
}


sint4 SpecialWorker::getCurrentMinerals() const{ 
    return gob->get_int("minerals"); 
}


GameObj* SpecialWorker::get_gob() const{
    return gob;
}


TerrainBase::Loc SpecialWorker::getLocation() const{
    return TerrainBase::Loc(*(gob->sod.x),*(gob->sod.y)); 
}


const History::Record* SpecialWorker::getMineralRec() const{ 
    if(route != 0){
        return route->getMineralRec(); 
    }else{
        return NULL;
    }
}


RoutePtr SpecialWorker::getRoute() const{
    // PRECONDITION //
    assert( route != 0 );
    return route;
}


sint4 SpecialWorker::getState() const{
    return state;
}


bool SpecialWorker::hasRoute() const{    
	return route != 0;
}


bool SpecialWorker::isMining() const{
    return state == MINING;
}


void SpecialWorker::mine(){
    assert(route != 0);

    setState(MINING);
    PlayerInfo& pi = gsm->get_game().get_cplayer_info();
    Vector<sint4> args(1, pi.get_id(route->getMineralRec()->getObject()));
    gob->component(miningComponent)->set_action(miningAction, args);
}


void SpecialWorker::moveToCC(){
    assert(route != 0);

    route->moveToCC(gob);
    setState(RETURNING);
}


void SpecialWorker::moveToMineral(){
    assert(route != 0);

    setState(MOVING);
    route->moveToMineral(gob);
}


void SpecialWorker::setRoute(RoutePtr route){
    assert(route != NULL);

    if(this->route != 0){
        this->route->removeWorker();
    }

    this->route = route;
    this->route->addWorker();	
}


void SpecialWorker::setState(sint4 state_){
    state = state_;
}
